from dataclasses import dataclass
from enum import Enum
import json

import base64
import hashlib
import ssl
from typing import List, Optional
from urllib.parse import urlparse, urlunparse
import requests
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from .verify import Report, verify_attestation, CertificateChain

class PredicateType(str, Enum):
    """Predicate types for attestation"""
    AWS_NITRO_ENCLAVE_V1 = "https://tinfoil.sh/predicate/aws-nitro-enclave/v1"
    SEV_GUEST_V1 = "https://tinfoil.sh/predicate/sev-snp-guest/v1"

ATTESTATION_ENDPOINT = "/.well-known/tinfoil-attestation"

class AttestationError(Exception):
    """Base class for attestation errors"""
    pass

class FormatMismatchError(AttestationError):
    """Raised when attestation formats don't match"""
    pass

class MeasurementMismatchError(AttestationError):
    """Raised when measurements don't match"""
    pass

@dataclass
class Measurement:
    """Represents measurement data"""
    type: PredicateType
    registers: List[str]

    def fingerprint(self) -> str:
        """
        Computes the SHA-256 hash of all measurements, 
        or returns the single measurement if there is only one
        """
        if len(self.registers) == 1:
            return self.registers[0]

        all_data = str(self.type) + "".join(self.registers)
        return hashlib.sha256(all_data.encode()).hexdigest()

    def equals(self, other: 'Measurement') -> None:
        """
        Checks if this measurement equals another measurement
        Raises appropriate error if they don't match
        """
        if self.type != other.type:
            raise FormatMismatchError()
        if len(self.registers) != len(other.registers) or self.registers != other.registers:
            raise MeasurementMismatchError()

@dataclass
class Verification:
    """Represents verification results"""
    measurement: Measurement
    public_key_fp: str

@dataclass
class Document:
    """Represents an attestation document"""
    format: PredicateType
    body: str

    def hash(self) -> str:
        """Returns the SHA-256 hash of the attestation document"""
        all_data = str(self.format) + self.body
        return hashlib.sha256(all_data.encode()).hexdigest()

    def verify(self) -> Verification:
        """
        Checks the attestation document against its trust root 
        and returns the inner measurements
        """
        if self.format == PredicateType.SEV_GUEST_V1:
            return verify_sev_attestation(self.body)
        else:
            raise ValueError(f"Unsupported attestation format: {self.format}")

def verify_attestation_json(json_data: bytes) -> Verification:
    """Verifies an attestation document in JSON format and returns the inner measurements"""
    doc_dict = json.loads(json_data)
    doc = Document(
        format=PredicateType(doc_dict["format"]),
        body=doc_dict["body"]
    )
    return doc.verify()

def key_fp(public_key: ec.EllipticCurvePublicKey) -> str:
    """Returns the fingerprint of a given ECDSA public key"""
    key_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    return hashlib.sha256(key_bytes).hexdigest()

def cert_pubkey_fp(cert: x509.Certificate) -> str:
    """Returns the fingerprint of the public key of a given certificate"""
    pub_key = cert.public_key()
    if not isinstance(pub_key, ec.EllipticCurvePublicKey):
        raise ValueError(f"Unsupported public key type: {type(pub_key)}")
    
    return key_fp(pub_key)

def connection_cert_fp(ssl_socket: ssl.SSLSocket) -> str:
    """Gets the KeyFP of the public key of a TLS connection"""
    cert_bin = ssl_socket.getpeercert(binary_form=True)
    if not cert_bin:
        raise ValueError("No peer certificates")
    
    cert = x509.load_der_x509_certificate(cert_bin)
    return cert_pubkey_fp(cert)

def fetch_attestation(host: str) -> Document:
    """Retrieves the attestation document from a given enclave hostname"""
    url = f"https://{host}{ATTESTATION_ENDPOINT}"
    response = requests.get(url)
    response.raise_for_status()
    
    doc_dict = response.json()
    return Document(
        format=PredicateType(doc_dict["format"]),
        body=doc_dict["body"]
    )

def verify_sev_attestation(attestation_doc: str) -> Verification:
    """Verify SEV attestation document and return verification result."""
    try:
        att_doc_bytes = base64.b64decode(attestation_doc)
    except Exception as e:
        raise ValueError(f"Failed to decode base64: {e}")
    
    # Parse the report
    try:
        report = Report(att_doc_bytes)
    except Exception as e:
        raise ValueError(f"Failed to parse report: {e}")
    
    # Get attestation chain
    chain: CertificateChain = CertificateChain.from_report(report)

    # Verify attestation
    try:
        res = verify_attestation(chain, report)
    except Exception as e:
        raise ValueError(f"Failed to verify attestation: {e}")
    
    if res!= True:
        raise ValueError("Attestation verification failed!")

    # Create measurement object
    measurement = Measurement(
        type=PredicateType.SEV_GUEST_V1,
        registers=[
            report.measurement.hex()
        ]
    )

    # The public key fingerprint is at the start of the report (32 bytes)
    kfp = report.report_data.decode()

    return Verification(
        measurement=measurement,
        public_key_fp=kfp
    )


import http.client
import json
import ssl
import urllib.request
import socket
from dataclasses import dataclass
from typing import Dict, Optional, List
from urllib.parse import urlparse
import cryptography.x509
from cryptography.hazmat.primitives.serialization import PublicFormat, Encoding
import hashlib

from .attestation import fetch_attestation
from .github import fetch_latest_digest, fetch_attestation_bundle
from .sigstore import verify_attestation


@dataclass
class GroundTruth:
    """Represents the "known good" verified state of the enclave"""
    public_key: str  # Changed from cert_fingerprint to public_key
    digest: str
    measurement: str


class Response:
    """Represents an HTTP response"""
    def __init__(self, status: str, status_code: int, body: bytes):
        self.status = status
        self.status_code = status_code
        self.body = body


class TLSBoundHTTPSHandler(urllib.request.HTTPSHandler):
    """Custom HTTPS handler that verifies certificate public keys"""
    
    def __init__(self, expected_pubkey: str):
        super().__init__()
        self.expected_pubkey = expected_pubkey

    def https_open(self, req):
        return self.do_open(self._create_connection, req)

    def _create_connection(self, host, **kwargs):
        conn = super().do_open(http.client.HTTPSConnection, host, **kwargs)
        if not conn.sock:
            raise ValueError("No TLS connection")
        
        cert = conn.sock.getpeercert(binary_form=True)
        if not cert:
            raise ValueError("No valid certificate")
        
        public_key = cert.public_key()
        # Get the public key in PKIX/DER format
        public_key_der = public_key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)
        # Hash the public key
        cert_fp = hashlib.sha256(public_key_der).hexdigest()

        if cert_fp != self.expected_pubkey:
            raise ValueError(f"Certificate public key fingerprint mismatch: expected {self.expected_pubkey}, got {cert_fp}")
        
        return conn


class SecureClient:
    """A client that verifies and communicates with secure enclaves"""
    
    def __init__(self, enclave: str, repo: str):
        self.enclave = enclave
        self.repo = repo
        self._ground_truth: Optional[GroundTruth] = None

    @property
    def ground_truth(self) -> Optional[GroundTruth]:
        """Returns the last verified enclave state"""
        return self._ground_truth

    def verify(self) -> GroundTruth:
        """
        Fetches the latest verification information from GitHub and Sigstore
        and stores the ground truth results in the client
        """
        try:
            # Fetch and verify all required information
            digest = fetch_latest_digest(self.repo)
            sigstore_bundle = fetch_attestation_bundle(self.repo, digest)
            
            code_measurements = verify_attestation(
                sigstore_bundle, 
                digest, 
                self.repo
            )
            
            enclave_attestation = fetch_attestation(self.enclave)
            verification = enclave_attestation.verify()
            
            # Verify measurements match
            for (i, code_measurement) in enumerate(code_measurements.registers):
                if code_measurement != verification.measurement.registers[i]:
                    raise ValueError("Code measurements do not match")

            # Build ground truth from the verified attestation
            self._ground_truth = GroundTruth(
                public_key=verification.public_key_fp,
                digest=digest,
                measurement=verification.measurement
            )
            return self._ground_truth
        
        except Exception as e:
            raise e

    def _get_http_client(self) -> urllib.request.OpenerDirector:
        """Returns an HTTP client that only accepts TLS connections to the verified enclave"""
        if not self._ground_truth:
            ground_truth, err = self.verify()
            if err:
                raise err
            if not ground_truth:
                raise ValueError("Failed to verify enclave")
        
        handler = TLSBoundHTTPSHandler(self._ground_truth.public_key)
        return urllib.request.build_opener(handler)

    def _make_request(self, req: urllib.request.Request) -> Response:
        """Makes an HTTP request using the secure client"""
        client = self._get_http_client()
        
        # If URL doesn't have a host, assume it's relative to the enclave
        if not urlparse(req.full_url).netloc:
            req.full_url = f"https://{self.enclave}{req.full_url}"
        
        with client.open(req) as resp:
            return Response(
                status=f"{resp.status} {resp.reason}",
                status_code=resp.status,
                body=resp.read()
            )

    def post(self, url: str, headers: Dict[str, str], body: bytes) -> Response:
        """Makes an HTTP POST request"""
        req = urllib.request.Request(
            url,
            data=body,
            headers=headers,
            method="POST"
        )
        return self._make_request(req)

    def get(self, url: str, headers: Dict[str, str]) -> Response:
        """Makes an HTTP GET request"""
        req = urllib.request.Request(
            url,
            headers=headers,
            method="GET"
        )
        return self._make_request(req)



from .attestation import (
    fetch_attestation,
    verify_sev_attestation,
    Measurement,
    PredicateType
)

__all__ = [
    'fetch_attestation',
    'verify_sev_attestation',
    'Measurement',
    'PredicateType'
]